import uuid

# This is a static uuid used as seed for generating reproducible uuids based on
# text This is a temporary solution until we can get sources (uuids) from the
# ACT backend
ACT_UUID_NAMESPACE = uuid.UUID('6b159598-b585-11e7-a8e7-e83935122d71')
FACT_IS_SOURCE = "FactIsSource"
FACT_IS_DESTINATION = "FactIsDestination"
BIDIRECTIONAL_FACT = "BiDirectional"
DEFAULT_VALIDATOR = r'(.|\n)+'

RE_UUID_MATCH = r'^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$'
RE_UUID = r'(?P<uuid>[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12})'
RE_TIMESTAMP_MATCH = r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z$'
RE_TIMESTAMP = r'(?P<timestamp>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z)'

from act import utils
from act import schema
from act import base
from act import obj
from act import fact
from act import helpers

from .helpers import Act



import json
import copy
from logging import error
import requests
from .schema import Schema, Field, schema_doc


class NotImplemented(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ArgumentError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ResponseError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


def request(method, user_id, url, requests_common_kwargs = None, **kwargs):
    """Perform requests towards API

Args:
    method (str):         POST|GET
    user_id (int):        Act user ID
    url (str):            Absolute URL for the endpoint
    **kwargs (keywords):  Additional options passed to requests json parameter
                          the following fields:
"""

    if not requests_common_kwargs:
        requests_common_kwargs = {}

    requests_kwargs = copy.copy(requests_common_kwargs)

    # Apply common request arguments
    requests_kwargs.update(kwargs)

    res = requests.request(
        method,
        url,
        headers={
            "ACT-User-ID": str(user_id)
        },
        **requests_kwargs
    )

    if res.status_code == 412:
        error("Request failed: {}, {}, {}".format(
            url, kwargs, res.status_code))
        raise ResponseError(res.text)

    elif res.status_code not in (200, 201):
        error("Request failed: {}, {}, {}".format(
            url, kwargs, res.status_code))
        raise ResponseError(
            "Unknown response error {}: {}".format(res.status_code, res.text))

    try:
        return res.json()

    except json.decoder.JSONDecodeError:
        raise ResponseError(
            "Error decoding response {}: {}".format(res.status_code, res.text))


class ActResultSet(object):
    """Represents a list of Act entries"""

    def __init__(self, response, deserializer):
        """Initialize result set
Args:
    response (str):       JSON response from Act. This should include
                          the following fields:
                            - count: the number of entries fetched
                            - limit: the limit in the query
                            - responseCode: responseCode from the API
                            - size: the total number of entries in the platform
                            - data (array): array of entries

    deserializer (class): Deseralizer class
"""

        if not isinstance(response["data"], list):
            raise ResponseError(
                "Response should be list: {}".format(
                    response["data"]))

        self.data = [deserializer(**d) for d in response["data"]]

        self.size = response["size"]
        self.count = response["count"]
        self.limit = response["limit"]
        self.status_code = response["responseCode"]

    @property
    def complete(self):
        """Returns true if we have recieved all data that exists on the endpoint"""
        return self.size >= self.count

    def __call__(self, func, *args, **kwargs):
        """Call function on each data entry"""

        self.data = [getattr(item, func)(*args, **kwargs)
                     for item in self.data]
        return self

    def __len__(self):
        """Returns the number of entries"""
        return len(self.data)

    def __repr__(self):
        return str(self.data)

    def __getitem__(self, sliced):
        return self.data[sliced]

    def __iter__(self):
        """Iterate over the entries"""
        return self.data.__iter__()


class Config(object):
    """Config object"""
    act_baseurl = None
    user_id = None
    requests_common_kwargs = {}

    def __init__(
            self,
            act_baseurl,
            user_id,
            requests_common_kwargs = None):
        """Set URL and USER ID, and optionally other arguments that will be passed on to request"""

        self.act_baseurl = act_baseurl
        self.user_id = user_id
        self.requests_common_kwargs = requests_common_kwargs


class ActBase(Schema):
    """Act object inheriting Schema, to support serializing and
    deserializing."""

    config = None

    SCHEMA = []

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(ActBase, self).__init__(*args, **kwargs)

    def configure(self, config):
        """Set config object"""

        self.config = config
        return self

    def api_request(self, method, uri, **kwargs):
        """Send request to API and update current object with result"""

        response = request(
            method,
            self.config.user_id,
            "{}/{}".format(self.config.act_baseurl, uri),
            **self.config.requests_common_kwargs,
            **kwargs
        )

        return response

    def api_post(self, uri, **kwargs):
        """Send POST request to API with keywords as JSON arguments"""

        return self.api_request("POST", uri, json=kwargs)

    def api_put(self, uri, **kwargs):
        """Send PUT request to API with keywords as JSON arguments"""

        return self.api_request("PUT", uri, json=kwargs)

    def api_get(self, uri):
        """Send GET request to API"""

        return self.api_request("GET", uri)


class NameSpace(ActBase):
    """Namespace - serialized object specifying Namespace"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]


class Organization(ActBase):
    """Manage FactSource"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]


class Source(ActBase):
    """Manage FactSource"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]


class Comment(ActBase):
    """Namespace - serialized object specifying Namespace"""

    SCHEMA = [
        Field("comment"),
        Field("id"),
        Field("timestamp", serializer=False),
        Field("reply_to"),
        Field("source", deserializer=Source, serializer=False),
    ]



from logging import info
import re
import json
import time
import act
from act import RE_UUID_MATCH

from .schema import Field, schema_doc, MissingField, ValidationError
from .obj import Object, ObjectType, object_binding_serializer
from .base import ActBase, Organization, NameSpace, Comment


class ObjectBinding(ActBase):
    SCHEMA = [
        Field(
            "object_type",
            deserializer=ObjectType,
            serializer=lambda object_type: object_type.id),
        Field("direction"),
    ]


class FactType(ActBase):
    """Manage FactType"""
    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("validator", default="RegexValidator"),
        Field("validator_parameter", default=act.DEFAULT_VALIDATOR),
        Field("entity_handler", default="IdentityHandler"),
        Field("entity_handler_parameter"),
        Field("relevant_object_bindings", deserializer=ObjectBinding),
        Field("namespace", deserializer=NameSpace),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(FactType, self).__init__(*args, **kwargs)

    def add(self):
        params = self.serialize()
        fact_type = self.api_post("v1/factType", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**fact_type)

        info("Created fact type: {}".format(self.name))

        return self

    def add_binding(self, direction, object_type):
        """Add binding"""
        if not self.id:
            raise MissingField("Must have fact type ID")

        url = "v1/factType/uuid/{}".format(self.id)

        add_object_bindings = [{
            "objectType": object_type,
            "direction": direction
        }]

        fact_type = self.api_put(
            url, addObjectBindings=add_object_bindings)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        info("Added binding to fact type {}: {} ({})".format(
            self.id, object_type, direction))

        return self

    def add_source_binding(self, object_type):
        return self.add_binding(act.FACT_IS_DESTINATION, object_type)

    def add_destination_binding(self, object_type):
        return self.add_binding(act.FACT_IS_SOURCE, object_type)

    def add_bidirectional_binding(self, object_type):
        return self.add_binding(act.BIDIRECTIONAL_FACT, object_type)

    def rename(self, name):
        if not self.id:
            raise MissingField("Must have fact type ID")

        old_name = self.name

        url = "v1/factType/uuid/{}".format(self.id)
        fact_type = self.api_put(url, name=name)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        info("Renamed fact type {}: {} -> {}".format(self.id, old_name, name))

        return self


class RetractedFact(ActBase):
    """Retracted Fact"""

    SCHEMA = [
        Field("type", deserializer=FactType,
              serializer=lambda fact_type: fact_type.name),
        Field("value", default="-"),
        Field("id", serializer=False),
    ]


class Fact(ActBase):
    """Manage facts"""

    SCHEMA = [
        Field("type", deserializer=FactType,
              serializer=lambda fact_type: fact_type.name),
        Field("value", default="-"),
        Field("id", serializer=False),
        # For now, do not serialize/deserialize source
        Field("source", deserializer=False, serializer=False),
        Field("timestamp", serializer=False),
        Field("last_seen_timestamp", serializer=False),
        Field("in_reference_to", deserializer=RetractedFact, serializer=False),
        Field("organization", deserializer=Organization, serializer=False),
        Field("access_mode", default="Public"),
        Field("objects", default=[], serialize_target="bindings",
              deserializer=Object, serializer=object_binding_serializer),
        Field("bindings", deserialize_target="objects", serializer=False),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(Fact, self).__init__(*args, **kwargs)

    def __add_if_not_exists(self, *args, **kwargs):
        """Add binding if it does not exist"""
        binding = Object(*args, **kwargs)
        if binding not in self.data["objects"]:
            self.data["objects"].append(binding)

        return self

    def source(self, *args, **kwargs):
        """Add source binding. Takes all arguments applicable to
        act.Act.object"""

        return self.__add_if_not_exists(
            *args, direction=act.FACT_IS_DESTINATION, **kwargs)

    def destination(self, *args, **kwargs):
        """Add destination binding. Takes all arguments applicable to
        act.Act.object"""

        return self.__add_if_not_exists(
            *args, direction=act.FACT_IS_SOURCE, **kwargs)

    def bidirectional(self, *args, **kwargs):
        """Add bidirectional binding. Takes all arguments applicable to act.Act.object"""
        return self.__add_if_not_exists(
            *args, direction=act.BIDIRECTIONAL_FACT, **kwargs)

    def get(self):
        """Get fact"""

        if not self.id:
            raise MissingField(
                "Must have fact ID to get, or use search instead")

        fact = self.api_get("v1/fact/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**fact)
        return self

    def add(self):
        """Add fact"""

        started = time.time()

        params = self.serialize()
        fact = self.api_post("v1/fact", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**fact)

        info("Created fact in %.2fs: data=%s" %
             (time.time() - started, json.dumps(fact)))

        return self

    def get_acl(self):
        """Get acl"""

        if not self.id:
            raise MissingField("Must have fact ID to resolve ACL")

        return self.api_get("v1/fact/uuid/{}/access".format(self.id))["data"]

    def grant_access(self, subject_uuid):
        """Grant access - not implemented yet"""

        raise act.base.NotImplemented("Grant access is not implemented")

    def get_comments(self):
        """Get comments
Args:
    None

Returns ActResultSet of Comments.

        """

        if not self.id:
            raise MissingField("Must have fact ID to get comments")

        res = self.api_get("v1/fact/uuid/{}/comments".format(self.id))
        return act.base.ActResultSet(res, Comment)

    def add_comment(self, comment, reply_to=None):
        """Add comment
Args:
    comment (str):    Comment to add
    reply_to (uuid):  Comment is a reply to another comment (specified by uuid)

Returns Fact object
        """

        if not self.id:
            raise MissingField("Must have fact ID to add comment")

        if reply_to and not re.search(RE_UUID_MATCH, reply_to):
            raise ValidationError("reply_to is not a valid UUID: {}".format(reply_to))

        self.api_post(
            "v1/fact/uuid/{}/comments".format(self.id),
            comment=comment, replyTo=reply_to)

        return self

    # pylint: disable=unused-argument,dangerous-default-value
    def retract(
            self,
            organization=None,
            source=None,
            access_mode=None,
            comment=None,
            acl=[]):
        """Retract fact
Args:
    organization (str): Set owner of new Fact. If not set the current user's
                        organization will be used (takes Organization UUID)
    source (str):       Set Source of new Fact. If not set the current user
                        will be used as Source (takes Source UUID)
    access_mode:        Set access mode of new Fact. If not set the accessMode
                        from the retracted Fact will be used = ['Public',
                        'RoleBased', 'Explicit']
    comment (str):      If set adds a comment to new Fact
    acl (str[] | str):  If set defines explicitly who has access to new Fact (takes Subject UUIDs)

All arguments are optional.

Returns retracted fact.
    """

        params = act.utils.prepare_params(locals(), ensure_list=["acl"])

        if self.id:
            url = "v1/fact/uuid/{}/retract".format(self.id)
        else:
            raise MissingField(
                "Must have object ID to retract object")

        fact = self.api_post(url, **params)["data"]

        self.data = {}
        self.deserialize(**fact)

        return self



import copy
import json
from .utils import snake_to_camel, camel_to_snake


def default_serializer(value):
    """Default serializer. return value without modification"""
    return value


class ValidationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class MissingField(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class Field(object):
    """Schema fields"""

    def __init__(
            self,
            name,
            default=None,
            serializer=None,
            deserializer=default_serializer,
            serialize_target=None,
            deserialize_target=None,
            flatten=None):
        """
Args:
    name (str):                Field name
    default (str):             Default value for field if not specified
    serializer (func|False):   Function used to serialize the field. Set to
                               False if field should not be serialized.
    deserializer (func|False): Function used to deserialize data. Set to
                               False if field should not be deserialized.
    serialize_target (str):    Move data to this key(name) when erializing
    deserialize_target (str):  Move data to this key(name) when deserializing
    """
        self.name = name
        self.default = default
        self.serializer = serializer
        self.deserializer = deserializer
        self.serialize_target = serialize_target
        self.deserialize_target = deserialize_target
        self.flatten = flatten

    def deserialize(self, value):
        """Deserialize field using suplied value"""

        # If value is already a Schema object, use it as is
        if isinstance(value, Schema):
            return value

        if isinstance(value, dict):
            if self.deserializer == default_serializer:
                raise ValidationError(
                    "dict is not suppored by default serializer. field={}, value={}".format(
                        self.name, value))

            return self.deserializer(**value)

        return self.deserializer(value)

    def serialize(self, value):
        """Serialize field using suplied value"""

        # Custom serializer
        if self.serializer:
            return self.serializer(value)

        # Nested serializer
        if hasattr(value, "serialize") and callable(value.serialize):
            return value.serialize()

        return value


def schema_doc(schema):
    """Dynamic create doctstring from schema fields"""
    def dec(obj):
        docstring = obj.__doc__ or ""
        dynamic = """

The following arguments can be specified by position or keyword.
    {}
""".format(str("\n    ".join([field.name for field in schema])))

        obj.__doc__ = docstring + dynamic

        return obj
    return dec


class Schema(object):
    """Define schema and serialize/deserialize response from the platform"""

    def __init__(self, *args, **kwargs):
        """Initialize (deserialize) fields"""

        self.data = {}

        # Apply all positinal arguments based on order in SCHEMA
        for i, v in enumerate(args):
            kwargs[self.SCHEMA[i].name] = v

        self.deserialize(**kwargs)

    def json(self, exclude_empty=True, to_camel_case=True):
        return json.dumps(self.serialize(exclude_empty=exclude_empty, to_camel_case=to_camel_case))

    def serialize(self, exclude_empty=True, to_camel_case=True):
        entries = {}
        for field in self.SCHEMA:
            serialize_target = field.serialize_target or field.name

            if to_camel_case:
                serialize_target = snake_to_camel(serialize_target)

            # Entry is stored in another field internally
            value = self.data.get(field.name, None)

            # Exclude empty values
            if exclude_empty and not value:
                continue

            # Serializer disabled
            if field.serializer is False:
                continue

            if isinstance(value, (list, tuple)):
                entries[serialize_target] = [field.serialize(v) for v in value]
            else:
                entries[serialize_target] = field.serialize(value)

        return entries

    def get_field(self, name):
        for field in self.SCHEMA:
            if name == field.name:
                return field
        return None

    def get_deserialize_field(self, name):
        field = self.get_field(name)

        if not field:
            return None

        if field.deserialize_target:
            return self.get_field(field.deserialize_target)
        return field

    def get_serialize_field(self, name):
        field = self.get_field(name)

        if not field:
            return None

        if field.serialize_target:
            return self.get_field(field.serialize_target)

        return field

    def deserialize(self, **entries):
        for k, value in entries.items():
            k = camel_to_snake(k)
            field = self.get_deserialize_field(k)

            if not field:
                raise ValidationError(
                    '"{}" not defined in schema on {}'.format(
                        k, self.__class__))

            # deserializer disabled
            if field.deserializer is False:
                continue

            if field.flatten:
                self.deserialize(**value)
                continue

            if isinstance(value, (list, tuple)):
                self.data[field.name] = [field.deserialize(v) for v in value]
            else:
                self.data[field.name] = field.deserialize(value)

        if not hasattr(self, "SCHEMA"):
            raise ValidationError(
                "No SCHEMA defined in class {}".format(self.__class__))

        for field in self.SCHEMA:
            # Loop through all fields and set default value, unless any of
            # - field is deseriazlied with another key
            # - field is flattened
            # - deserializer is disabled
            if field.name not in self.data \
                    and not field.deserialize_target \
                    and not field.flatten \
                    and field.deserializer is not False:
                self.data[field.name] = copy.copy(field.default)

    def __getattr__(self, attr):
        if attr in self.data:
            return self.data[attr]
        else:
            raise AttributeError(
                # pylint: disable=too-many-format-args
                "{} object has no attribute {}".format(
                    self.__class__, attr))

    def __repr__(self):
        return repr(self.data)



import re
import responses
import act
from act import RE_UUID_MATCH, RE_TIMESTAMP_MATCH, RE_UUID, RE_TIMESTAMP
from act_test import get_mock_data

# pylint: disable=no-member

@responses.activate
def test_add_fact():
    mock = get_mock_data("data/post_v1_fact_201.json")
    responses.add(
        responses.POST,
        mock["url"],
        json=mock["json"],
        status=mock["status_code"])

    c = act.Act("http://localhost:8080", 1)

    f = c.fact("seenIn", "report") \
        .source("ipv4", "127.0.0.1") \
        .destination("report", "xyz")

    assert f.type.name == "seenIn"
    assert f.value == "report"

    # We do not have id or timestamp yet
    assert f.id is None
    assert f.timestamp is None
    assert len(f.objects) == 2

    # Add fact
    f.add()

    # id, timestamp and organization should now be fetched from API
    assert re.search(RE_UUID_MATCH, f.id)
    assert re.search(RE_TIMESTAMP_MATCH, f.timestamp)
    assert re.search(RE_UUID_MATCH, f.organization.id)
    # Not implemented/stable in backend API yet
    # self.assertRegex(f.origin.id, RE_UUID_MATCH)
    assert len(f.objects) == 2


@responses.activate
def test_get_fact_types():
    mock = get_mock_data("data/get_v1_factType_200.json")
    responses.add(
        responses.GET,
        mock["url"],
        json=mock["json"],
        status=mock["status_code"])

    c = act.Act("http://localhost:8080", 1)
    fact_types = c.get_fact_types()

    assert fact_types.size == len(fact_types)


@responses.activate
def test_fact_search():
    mock = get_mock_data("data/post_v1_fact_search_200.json")
    responses.add(
        responses.POST,
        mock["url"],
        json=mock["json"],
        status=mock["status_code"])

    c = act.Act("http://localhost:8080", 1)

    facts = c.fact_search(
        fact_type=["seenIn"],
        fact_value=["report"],
        limit=25)

    assert not facts.complete
    assert facts.size == 25
    assert facts.count > 1000

    # All facts should have an UUID
    assert all([re.search(RE_UUID_MATCH, fact.id) for fact in facts])


@responses.activate
def test_fact_acl():
    mock = get_mock_data("data/get_v1_fact_uuid_access_200.json")
    responses.add(
        responses.GET,
        mock["url"],
        json=mock["json"],
        status=mock["status_code"])

    c = act.Act("http://localhost:8080", 1)
    uuid = re.search(RE_UUID, mock["url"]).group("uuid")
    acl = c.fact(id=uuid).get_acl()
    assert acl == []


@responses.activate
def test_fact_get_comments():
    mock = get_mock_data("data/get_v1_fact_uuid_comments_200.json")
    responses.add(
        responses.GET,
        mock["url"],
        json=mock["json"],
        status=mock["status_code"])

    c = act.Act("http://localhost:8080", 1)

    # Get comments
    uuid = re.search(RE_UUID, mock["url"]).group("uuid")
    comments = c.fact(id=uuid).get_comments()
    assert comments  # Should be non empty
    assert comments[0].comment == "Test comment"
    assert re.search(RE_TIMESTAMP, comments[0].timestamp)
    assert all([isinstance(comment, act.base.Comment) for comment in comments])


@responses.activate
def test_fact_add_comment():
    mock = get_mock_data("data/post_v1_fact_uuid_comments_201.json")
    responses.add(
        responses.POST,
        mock["url"],
        json=mock["json"],
        status=mock["status_code"])

    c = act.Act("http://localhost:8080", 1)

    # Get comments
    uuid = re.search(RE_UUID, mock["url"]).group("uuid")

    c.fact(id=uuid).add_comment("Test comment")



from logging import info, warning
import act
from .schema import Field, schema_doc, MissingField
from .base import ActBase, NameSpace


class ObjectType(ActBase):
    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("validator", default="RegexValidator"),
        Field("validator_parameter", default=act.DEFAULT_VALIDATOR),
        Field("entity_handler", default="IdentityHandler"),
        Field("entity_handler_parameter"),
        Field("namespace", deserializer=NameSpace),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(ObjectType, self).__init__(*args, **kwargs)

    def add(self):
        params = self.serialize()
        object_type = self.api_post("v1/objectType", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**object_type)

        info("Created object type: {}".format(self.name))

        return self


class ObjectStatistics(ActBase):
    """ObjectStatistics - serialized object specifying statistics"""

    SCHEMA = [
        Field("type", deserializer=ObjectType),
        Field("count"),
        Field("last_seen_timestamp", serializer=False),
        Field("last_added_timestamp", serializer=False),
    ]


class Object(ActBase):
    """Manage objects"""
    SCHEMA = [
        Field(
            "type",
            deserializer=ObjectType,
            serializer=lambda object_type: object_type.name),
        Field("value"),
        Field("id"),
        Field(
            "statistics",
            deserializer=ObjectStatistics,
            serializer=False),
        Field("object", flatten=True),
        Field("direction"),
        Field("object_type", deserialize_target="type", serializer=False),
        Field("object_value", deserialize_target="value", serializer=False),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(Object, self).__init__(*args, **kwargs)

    def facts(self):
        """Get facts"""

        if self.id:
            url = "v1/object/uuid/{}/facts".format(self.id)
        elif self.type.name and self.value:
            url = "v1/object/{}/{}/facts".format(self.type.name, self.value)
        else:
            raise MissingField(
                "Must have either object ID or object type/value to get facts")

        response = self.api_post(url)

        result_set = act.base.ActResultSet(response, act.fact.Fact)

        # Add authentication information to all facts
        return result_set("configure", self.config)

    def traverse(self, query=None):
        """Traverse from object"""

        if self.id:
            url = "v1/object/uuid/{}/traverse".format(self.id)
        elif self.type.name and self.value:
            url = "v1/object/{}/{}/traverse".format(self.type.name, self.value)
        else:
            raise MissingField(
                "Must have either object ID or object type/value to get facts")

        result = []
        for element in self.api_post(url, query=query)["data"]:
            if "objects" in element:
                result.append(act.fact.Fact(**element))
            elif "statistics" in element:
                result.append(act.fact.Object(**element))
            else:
                warning("Unable to guess element type: {}".format(element))
                result.append(element)

        # This returns the list as is and it is currently not deserialized
        # since it can return multiplel types. An improvement would be to
        # autodetect the types and deserialize accordingly

        return result


def object_binding_serializer(obj):
    binding = {
        "direction": obj.direction
    }

    if "id" in obj.data and obj.id:
        binding["objectID"] = obj.id
    else:
        binding["objectType"] = obj.type.name
        binding["objectValue"] = obj.value

    return binding



import logging
from logging import warning
import sys
import act
from .base import ActBase, Config
from .fact import Fact, FactType
from .obj import Object, ObjectType
from .schema import schema_doc


class Act(ActBase):
    """Act class exposing most of the ACT API"""

    SCHEMA = []

    def __init__(
            self,
            act_baseurl,
            user_id,
            log_level="debug",
            log_file=None,
            log_prefix="act",
            requests_common_kwargs=None):
        super(Act, self).__init__()

        self.configure(Config(act_baseurl, user_id, requests_common_kwargs))

        self.setup_logging(log_level, log_file, log_prefix)

    def setup_logging(
            self,
            log_level="debug",
            log_file=None,
            log_prefix="act"):
        numeric_level = getattr(logging, log_level.upper(), None)

        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % log_level)

        datefmt = "%Y-%m-%d %H:%M:%S"
        formatter = "[%(asctime)s] app=" + log_prefix + \
            " level=%(levelname)s msg=%(message)s"

        if log_file:
            logging.basicConfig(
                level=numeric_level,
                filename=log_file,
                format=formatter,
                datefmt=datefmt)
        else:
            logging.basicConfig(
                level=numeric_level,
                stream=sys.stdout,
                format=formatter,
                datefmt=datefmt)

    # pylint: disable=unused-argument,dangerous-default-value
    def fact_search(
            self,
            keywords="",
            object_type=[],
            fact_type=[],
            object_value=[],
            fact_value=[],
            organization=[],
            source=[],
            include_retracted=None,
            before=None,
            after=None,
            limit=None):
        """Search objects
Args:
    keywords (str):               Only return Facts matching a keyword query
    object_type (str[] | str):    Only return Facts with objects having a specific
                                  ObjectType
    fact_type (str[] | str):      Only return Facts having a specific FactType
    object_value (str[] | str):   Only return Facts with objects matching a specific
                                  value
    fact_value (str[] | str):     Only return Facts matching a specific value
    organization (str[] | str):   Only return Facts belonging to
                                  a specific Organization
    source (str[] | str):         Only return Facts coming from a specific Source
    include_retracted (bool):     Include retracted Facts (default=False)
    before (timestamp):           Only return Facts added before a specific
                                  timestamp. Timestamp is on this format:
                                  2016-09-28T21:26:22Z
    after (timestamp):            Only return Facts added after a specific
                                  timestamp. Timestamp is on this format:
                                  2016-09-28T21:26:22Z
    limit (integer):              Limit the number of returned Objects
                                  (default 25, 0 means all)

All arguments are optional.

Returns ActResultSet of Facts.
    """

        params = act.utils.prepare_params(
            locals(),
            ensure_list=[
                "object_type",
                "object_value",
                "fact_type",
                "fact_value",
                "organization",
                "source"])

        res = self.api_post("v1/fact/search", **params)

        return act.base.ActResultSet(res, self.fact)

    # pylint: disable=unused-argument,dangerous-default-value
    def object_search(
            self,
            keywords="",
            object_type=[],
            fact_type=[],
            object_value=[],
            fact_value=[],
            organization=[],
            source=[],
            before=None,
            after=None,
            limit=None):
        """Search objects
Args:
    keywords (str):               Only return Objects matching a keyword query
    object_type (str[] | str):    Only return Objects with a specific ObjectType
    fact_type (str[] | str):      Only return Objects with Facts having a
                                  specific FactType
    object_value (str[] | str):   Only return Objects matching a specific
                                  value
    fact_value (str[] | str):     Only return Objects with Facts matching a
                                  specific value
    organization (str[] | str):   Only return Objects with Facts belonging to
                                  a specific Organization
    source (str[] | str):         Only return Objects with Facts coming from a
                                  specific Source
    before (timestamp):           Only return Objects with Facts added before
                                  a specific timestamp. Timestamp is on
                                  this format: 2016-09-28T21:26:22Z
    after (timestamp):            Only return Objects with Facts added after a
                                  specific timestamp. Timestamp is on
                                  this format: 2016-09-28T21:26:22Z
    limit (integer):              Limit the number of returned Objects
                                  (default 25, 0 means all)

All arguments are optional.

Returns ActResultSet of Objects.
    """

        params = act.utils.prepare_params(
            locals(),
            ensure_list=[
                "object_type",
                "object_value",
                "fact_type",
                "fact_value",
                "organization",
                "source"])

        res = self.api_post("v1/object/search", **params)

        return act.base.ActResultSet(res, self.object)

    @schema_doc(Fact.SCHEMA)
    def fact(self, *args, **kwargs):
        """Manage facts. All arguments are passed to create a Fact
object and authentication information is passed from the
act object."""

        return Fact(*args, **kwargs).configure(self.config)

    @schema_doc(Object.SCHEMA)
    def object(self, *args, **kwargs):
        """Manage objects. All arguments are passed to create an Object
object and authentication information is passed from the
act object."""

        return Object(*args, **kwargs).configure(self.config)

    @schema_doc(FactType.SCHEMA)
    def fact_type(self, *args, **kwargs):
        """Manage Fact Types. All arguments are passed to create a FactType
object and authentication information is passed from the
act object."""

        return FactType(*args, **kwargs).configure(self.config)

    @schema_doc(ObjectType.SCHEMA)
    def object_type(self, *args, **kwargs):
        """Manage Object Types. All arguments are passed to create a ObjectType
object and authentication information is passed from the
act object."""

        return ObjectType(*args, **kwargs).configure(self.config)

    def get_fact_types(self):
        """Get fact types"""

        return act.base.ActResultSet(
            self.api_get("v1/factType"), self.fact_type)

    def get_object_types(self):
        """Get object types"""

        return act.base.ActResultSet(
            self.api_get("v1/objectType"),
            self.object_type)

    def create_fact_type(
            self,
            name,
            validator=act.DEFAULT_VALIDATOR,
            source_objects=[],
            destination_objects=[],
            bidirectional_objects=[]):
        """Created fact type with given source, destination and bidirectional objects
Args:
    name (str):                  Fact type name
    validator (str):             Regular expression valdiator. Default = %s
    source_objects (str[]):      List of source objects (by name) linked to fact
    destination_objects (str[]): List of source objects (by name) linked to fact
    bidirection_objects (str[]): List of source objects (by name) linked to fact

Returns created fact type, or exisiting fact type if it already exists.
""" % act.DEFAULT_VALIDATOR

        existing_fact_types = {fact_type.name: fact_type
                               for fact_type in self.get_fact_types()}
        object_types = {
            object_type.name: object_type.id
            for object_type in self.get_object_types()}

        if name in existing_fact_types:
            warning("Fact type %s already exists" % name)
            return existing_fact_types[name]

        for obj in bidirectional_objects + source_objects + destination_objects:
            if obj not in object_types:
                raise act.base.ArgumentError(
                    "Illegal bidiectional object type {} linked from fact type {}".format(
                        obj, name))

        bindings = []

        for object_type_name in bidirectional_objects:
            bindings.append(
                {"objectType": {"id": object_types[object_type_name]},
                 "direction": act.BIDIRECTIONAL_FACT})

        for object_type_name in source_objects:
            bindings.append(
                {"objectType": {"id": object_types[object_type_name]},
                 "direction": act.FACT_IS_DESTINATION})

        for object_type_name in destination_objects:
            bindings.append(
                {"objectType": {"id": object_types[object_type_name]},
                 "direction": act.FACT_IS_SOURCE})

        return self.fact_type(
            name=name,
            validator_parameter=validator,
            relevant_object_bindings=bindings).add()

    def create_fact_type_all_bindings(
            self, name, validator_parameter=act.DEFAULT_VALIDATOR):
        """Create a fact type that can be connected to all object types"""

        existing_fact_types = {fact_type.name: fact_type
                               for fact_type in self.get_fact_types()}

        if name in existing_fact_types:
            warning("Fact type %s already exists" % name)
            return existing_fact_types[name]

        objectBindings = []

        for direction in [
                act.FACT_IS_SOURCE,
                act.FACT_IS_DESTINATION,
                act.BIDIRECTIONAL_FACT]:
            for object_type in self.get_object_types():
                objectBindings.append({
                    "objectType": {"id": object_type.id},
                    "direction": direction
                })

        return self.fact_type(name=name,
                              validator_parameter=validator_parameter,
                              relevant_object_bindings=objectBindings).add()
